package model;

/** 
 *This class provides methods to compute distances between a Farmer f and a LandUsePattern lp.
 *The complete threshold computed in "public double distance(LandUsePattern l, Farmer f)" 
 *and each of its subcomponents are computed as value in [0,1], weights can be used
* to adjust importance of subcomponents.
 * 
 * "distances" reflect various aspects that prevent adoption of a LandUsePattern lp by Farmer f:
 * - difference in planting of crops and usage of technology between f.landUsePattern and lp reflecting learning effort / uncertainty (Rogers' complexity)
 * - the capital / investments involved (e.g. planting trees takes time until profit can be made, 
 *  cutting trees potentially destroys capital; irrigation technology and machinery have to be written off 
 * 
 * Implementation:
 * - Learning efforts: Distance of a Farmer to a Crop or Technology is computed as minimum learning-distance to any
 * things currently used by this farmer. Learning-distances between crops and technologies are defined in a matrice.
 * -Capital: distance of a LandUse to a farmer is calculated as the capital-distances of different 
 *  crops and technologies weighted by the area dedicated to these Options. Capital-distances are defined
 *  in vectors.
 * 
 *  
 *  
 *  Some comments on distances between irrigation technologies:
   	- drip is more complex and less triable than sprinkler
  	- going back from sprinkler or drip to flooding means loss of capital
	- changing from sprinkler to drip or vice versa means at the same time loss of captial and investments
	
	Some notes on the capital-distance between crops:
	- costs refer to changing one ha to another crop class (reduction or increment in area are not distinguished
	  The following classes are distinguished: field crops, horticultural crops, vineyards, olives. 
	- field crops need machinery, horticultural crops mainly hand work (->less machinery to write off), tree crops 
	involve the trees themselves as well as certain machinery(?)
*
*/  

import uchicago.src.sim.util.Random;


public class Distances {
	
	public static final int SKILLS = 0;
	public static final int CAPITAL = 1;

	//in distance "skills": weight of tech compared to crop
	public static double skillsTechWeight = 0.5; 
	//in distance "capital": weight of tech compared to crop
	public static double capitalTechWeight = 0.5;
	
	//  Mean and standard deviation for normally distributed situation specific factor
	//	sd = mean/3.0 --> 99.7% of all random numbers are within [mean-3*sd,mean+3*sd]
	public static double situationSpecific_mean = 1.0;
	public static double situationSpecific_sd = 0.5/3.0;
	protected static double situationSpecific_upperBound = 2.0;//situationSpecific_mean+3.0*situationSpecific_sd;
	protected static double situationSpecific_lowerBound = 0.0;//situationSpecific_mean-3.0*situationSpecific_sd;
	
	
	
	/*********************************************/	
	/*********************************************/
	/*Distance related to skills and learning efforts*/
	/*********************************************/
	
	
	
	/** This methods summarizes uncertainties and necessary learning efforts 
	 * associated with a shift to l, arising from changes in crops and irrigation technologies.*/
	public static double getDistanceSkills(LandUsePattern lp, Farmer f){
		
		//Calculate total learning efforts
		
	
		
//*debug*/	System.out.println("Uncertainty: crops:"+d_crops+", tech:"+d_techs);
		
		//Calculate total learning/uncertainty distance considering learningTechWeight and normalize to [0,1]
		double dCrops =distanceSkillsCrops(lp,f);
		double dTechs = distanceSkillsTechs(lp,f);
			
		double d = (dCrops +skillsTechWeight*dTechs);///(1.0+skillsTechWeight);
		
		return d;
		
	}
	
	public static double distanceSkillsCrops(LandUsePattern lp, Farmer f){
	
//		distances defined for classes: change between crops within one class or across classes
		double [][] distancesCrops = 
		  {//0   1   2   3   4  //cf. numbers of crop classes as defined in cropClassFor(Crop c)
			
			{0.1, 0.3, 1.0, 1.0}, //0:COP -> sth.
			{0.3, 0.2, 1.0, 1.0},  //1:OFC -> sth.
			{0.5, 0.5, 0.5, 1.0},	//2:horticulture -> sth.
			{0.5, 0.5, 1.0, 1.0}  //3: tree crop-> sth.
		  };	
		
		//		Calculate learning efforts and uncertainties according to above tables
		
		//Summarize crops present in lp
		double lpCrops[] = new double[11];  //to sum up area of crops in lp; all fields 0 by default
	
		LandUse l;
		Farming o;
		for(int i=0; i< lp.length(); i++){
			l = lp.get(i);
			o = (Farming) (l.option);
			lpCrops[o.crop.type]+=l.size;
		}
		
		//Calculate learning efforts involved for Farmer f changing to LandUsePattern lp
		//Learning for new crops
		double dSum_crop =0.0;
		//for each crop of type i in new lp with size>0
		for(int i= 0; i<lpCrops.length; i++){
			if(lpCrops[i]>0.0){	
				//identify learning effort by comparing with farmers' experience (stored in f.knownCrops)
				if(!f.knowsCrop(i)){	//if farmer does not know Crop
					double dMin_crops=1.0;	//distance for a farmer without experience is 1.0
					double dc;
					//check for most similar crop the farmer has experience with 
					for(int j=0; j< f.cropSkills.length; j++){
						if((j!=Crop.FALLOW) && (f.knowsCrop(j))){	//Exception: Fallow does not provide experience with COP
							dc = distancesCrops[Crop.cropClassFor(j)][Crop.cropClassFor(i)]*(1+(1-f.cropSkill(j)));
							if(dc<dMin_crops){
								dMin_crops = dc;
							}
						}
					}
					dSum_crop += dMin_crops*lpCrops[i];
				}
				else{ //if farmer knows crop but extends area this also induces uncertainty
					double d = Math.max(0.0,lpCrops[i]-f.evaluatedLandUsePattern().pattern.cropArea(i));
					dSum_crop+=d*(1-f.cropSkill(i));
				}
			}	
		}

		return dSum_crop/f.sizeOfFarm;
	}
	
	
	
	public static double distanceSkillsTechs(LandUsePattern lp, Farmer f){
		
		/* Order of irrigation technologies in table:
		 *  0 = raindfed
		 *  1=floodIrrigation, 
		 *  2=sprinkler, 
		 *  3=drip 
		Table reads as follows distanceTechs[2][1] means threshold to switch
		from sprinkler to floodIrrigation.*/
		double [][] distancesTechs = 
		  {//0   1   2   3//cf. numbers of techs above
			{0.0, 0.5, 0.8, 1.0},  //0: rainfed --> sth.
			{0.0, 0.0, 0.3, 0.5 }, //1:floodIrrigation -> sth.
			{0.0, 0.0, 0.0, 0.3},  //2:sprinkler -> sth.
			{0.0, 0.0, 0.2, 0.0}  //3:drip -> sth.
					
		  };
		
		
		//		Summarize techs present in lp
		double lpTechs[] = new double[4];	//to sum up area of techs in lp; all fields 0 by default
			
		LandUse l;
		Farming o;
		for(int i=0; i< lp.length(); i++){
			l = lp.get(i);
			o = (Farming) (l.option);
			lpTechs[o.tech.type]+=l.size;
		}
		
		//Learning for new techs
		double dSum_tech =0.0;
		for(int i= 0; i<lpTechs.length; i++){
			if(lpTechs[i]>0.0){	//for each tech in l
//				identify learning effort by comparing with farmers' experience (stored in f.knownCrops)
				if(!f.knowsTech(i)){	//if farmer does not know Tech 
					double dMin_techs=Double.MAX_VALUE;
					double dt;
					//check for most similar tech the farmer has experience with 
					for(int j=0; j< IrrigationTechnology.NR_TECHS; j++){
						if(f.knowsTech(j)){
							dt = distancesTechs[j][i]*(1.0+(1.0-f.techSkill(j)));
							if(dt<dMin_techs){
								dMin_techs = dt;
							}
						}
					}
					dSum_tech += dMin_techs*lpTechs[i];
				}
				else{
					//if farmer knows tech but extends area this also induces uncertainty
					double d = Math.max(0.0,lpTechs[i]-f.evaluatedLandUsePattern().pattern.irrigationTechArea(i));
					dSum_tech+=d*(1-f.techSkill(i));
				}
			}
		}
		
		//normalize dSum_tech to [0,1]
		return dSum_tech/f.sizeOfFarm;
			
	}
	/*********************************************/	
	/*********************************************/
	/*Distance related to capital loss ***********/
	/*********************************************/
	
	
	public static double getDistanceCapital(LandUsePattern lp, Farmer f){
		
		double [][] reallocationOfCrops = reallocationCrops(lp,f);
		
		double lossCrops = distanceCapitalCrops(reallocationOfCrops,lp,f);
		double lossTechs = distanceCapitalTechs(lp,f);
		double loss = (lossCrops+capitalTechWeight*lossTechs); ///(1.0+capitalTechWeight);
		
		double situationSpecific;
		do{
			situationSpecific = Random.normal.nextDouble(situationSpecific_mean, situationSpecific_sd);
		}while((situationSpecific > situationSpecific_upperBound) || (situationSpecific < situationSpecific_lowerBound)); 
		
		loss = loss*situationSpecific; 	//in average this means "/2.0"	
		
		return loss;
	}
	
	/** Return value is normalized.	 */
	public static double distanceCapitalCrops(double[][] reallocationCrops, LandUsePattern lp, Farmer f){
	
		//Capital loss involved in changing one ha from class line to class column. Cost for changes
		//within one class refer to costs arising for changes to another type within this class
		//(e.g. changing from TraditionalCereals to Sunflower, both of which are COP).
		//0=COP, 1=OFC, 2= horticultural crops,3= tree crops. 
		double costsCropChange [][] = {
				{0.1, 0.2, 0.5, 0.8},	//COP-> sth.
				{0.2, 0.1, 0.5, 0.8},	//OFC -> sth.
				{0.5, 0.5, 0.2, 0.7},	//Horticulture -> sth.
				{0.8, 0.8, 0.7, 1.0}}; //Tree crops -> sth.

		
		
		//Costs for crop re-allocation (heuristic algorithm, depends on entries of costsCropChange [][]!		
		//compute costs related to change. 
	////
		//Compute costs arising from re-allocation according to re-allocation matrix
		double loss =0;
		for(int i=0; i<reallocationCrops.length; i++){
			for(int j=0; j<reallocationCrops[i].length; j++){
				loss += reallocationCrops[i][j]*costsCropChange[Crop.cropClassFor(i)][Crop.cropClassFor(j)];
			}
		}
		return loss/f.sizeOfFarm;
	}
	
	public static double distanceCapitalTechs(LandUsePattern lp, Farmer f){
		
		//Capital loss arises from a) not using sprinkler or drip equipment and b) from not
		//using full capacity of wells
		//CapitalLoss ~ 1 if sizeOfFarm is shifted from drip (sprinkler) to rainfed
		
		double sprinklerFactor=0.5, dripFactor=0.5, waterFactor=0.5;
		
		double dSprinkler = Math.abs(f.irrigationTech.getSprinklerArea()-lp.irrigationTechArea(IrrigationTechnology.SPRINKLER))/f.sizeOfFarm;
		double dDrip = Math.abs(f.irrigationTech.getDripArea()-lp.irrigationTechArea(IrrigationTechnology.DRIP))/f.sizeOfFarm;
		double dWater = Math.abs(f.irrigationTech.getWaterCapacity() - lp.waterUsed())/(f.sizeOfFarm*GlobalVariables.maxWaterUse);

//*debug*/	System.err.println("sprinkler: " +dSprinkler+", drip: "+dDrip+", water: "+dWater); 		
		
		return dSprinkler*sprinklerFactor+dDrip*dripFactor+dWater*waterFactor;
	}
	
	

	
	/*Question for implementation of this method: how to compute best re-allocation of area between different classes? The problem can be 
	//described as matrix multiplication AB whereas A contains in lines costs per changing 1 ha from to, and B 
	 * contains in columns area changed from X to something else (e.g. first column COP->) Constraints: 
	 * Sums of each line and column in B are fixed (lines= area increase in X; columns= area reduction 
	 * in X). Allocation of numbers in B with given constraints in a way that the sum of all entries in 
	 * matrice resulting from product AB is minimized. This might be bit complicated. An easier heuristic 
	 * solution using a matrix specific algorithm is implemented. 
	 */
	protected static double[][] reallocationCrops(LandUsePattern lp, Farmer f){
		
		double areasCropsLp[] = new double [11];		
		double areasCropsF[] = new double [11];
		double reallocation[][] = new double [11][11];
		
		LandUse l;
		Farming o;
		
		//inventory of f's current land-use
		for(int i=0; i<f.landUsePattern().length(); i++){
			l = f.landUsePattern().get(i);
			if(! (l.option instanceof Farming)) continue; //!! Only Farming options treated here!
			o = (Farming)l.option;
			areasCropsF[o.crop.type] += l.size;
		}
		//inventory of lp
		for(int i=0; i<lp.length(); i++){
			l = lp.get(i);
			if(! (l.option instanceof Farming)) continue;			//!! Only Farming options treated here!
			o = (Farming)l.option;
			areasCropsLp[o.crop.type] += l.size;
		}
		
		//remove all areas that remain same from areasCropsF and from areasCropsLp
		double a;
		for(int i=0; i<areasCropsF.length; i++){
			a=Math.min(areasCropsF[i], areasCropsLp[i]);
			areasCropsF[i]-=a;
			areasCropsLp[i]-=a;
		}

		//go through areasCropsF and re-allocate remaining areas
		//start with re-allocations within same crop class (except TREECROPS)
		for(int i=0; i<areasCropsF.length; i++){
			if(areasCropsF[i]>0.0){ //some area to re-allocate
				if(Crop.cropClassFor(i) == Crop.TREECROP) continue; //go to next entry in areasCropsF
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area in same crop class
					if((areasCropsLp[j]>0) && (Crop.cropClassFor(i) ==Crop.cropClassFor(j))){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		//re-allocations COP to OFC
		for(int i=0; i<areasCropsF.length; i++){
			if((areasCropsF[i]>0.0) && (Crop.cropClassFor(i)==Crop.COP)){ //some area to re-allocate
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area 
					if((areasCropsLp[j]>0) && (Crop.cropClassFor(j)==Crop.OFC)){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		//re-allocations OFC to COP
		for(int i=0; i<areasCropsF.length; i++){
			if((areasCropsF[i]>0.0) && (Crop.cropClassFor(i)==Crop.OFC)){ //some area to re-allocate
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area 
					if((areasCropsLp[j]>0) && (Crop.cropClassFor(j)==Crop.COP)){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		//re-allocations Horticulture to anything
		for(int i=0; i<areasCropsF.length; i++){
			if((areasCropsF[i]>0.0) && (Crop.cropClassFor(i)==Crop.HORTICULTURE)){ //some area to re-allocate
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area 
					if(areasCropsLp[j]>0){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		//re-allocations COP to anything left
		for(int i=0; i<areasCropsF.length; i++){
			if((areasCropsF[i]>0.0) && (Crop.cropClassFor(i)==Crop.COP)){ //some area to re-allocate
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area 
					if(areasCropsLp[j]>0){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		//re-allocations OFC to anything left
		for(int i=0; i<areasCropsF.length; i++){
			if((areasCropsF[i]>0.0) && (Crop.cropClassFor(i)==Crop.OFC)){ //some area to re-allocate
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area 
					if(areasCropsLp[j]>0){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		//re-allocations Tree crops (nothing else left)
		for(int i=0; i<areasCropsF.length; i++){
			if(areasCropsF[i]>0.0){ //some area to re-allocate
				for(int j=0; j<areasCropsLp.length; j++){ //look for "free" area 
					if(areasCropsLp[j]>0){
						a = Math.min(areasCropsF[i],areasCropsLp[j]);
						areasCropsF[i]-=a;
						areasCropsLp[j]-=a;
						reallocation[i][j] += a;
						if(areasCropsF[i]<=0.0) break; //all area of areasCropsF[i] re-allocated
					}
				}
			}	
		}
		return reallocation;
		
	}
}
