package model;

import java.util.ArrayList;
import java.util.ListIterator;

import model.decisionMaking.ChoiceProcess;
import model.decisionMaking.CollectOptions;
import model.decisionMaking.CollectOptions_DiffusionNrPeople;
import model.decisionMaking.CompareLandUsePattern;
import model.decisionMaking.CompareLandUsePattern_Satisficing;
import model.decisionMaking.EvaluatedLandUsePattern;
import model.decisionMaking.GenerateLandUsePattern;
import model.decisionMaking.RuleApplication;

import uchicago.src.sim.util.Random;


/**This class defines the farmer at the satisficing version of the Guadiana Model
 * 
 * @author Marvin Nebel (mnebel@uos.de)
 * @version 03/29/2011
 *
 */

public class Farmer_Satisficing extends Farmer {
	
	/* Thresholds */
	public static final int PROFIT = 0;
	public static final int LABOUR = 1;
	public static final int RISK = 2;
	public static final int LEGALITY = 3;
	
	
	/* Number of Thresholds */
	public static final int NUMBER_OF_THRESHOLDS = 4; 
	
	/* Thresholds included? */
	public static final boolean profitIncluded = PROFIT < NUMBER_OF_THRESHOLDS;
	public static final boolean labourIncluded = LABOUR < NUMBER_OF_THRESHOLDS;
	public static final boolean riskIncluded = RISK < NUMBER_OF_THRESHOLDS;
	public static final boolean legalityIncluded = LEGALITY < NUMBER_OF_THRESHOLDS;
	

	
	
	
	
	/* Number of past years that influence thresholds */
	public static final int DURATION_OF_HISTORY = 5;
	// Number of rounds, where Threshold will be decreased 
	// This number should be > 1000.
	// It is needed for debug reasons (if no satisfactory pattern can be found at all)
	public static int  MAX_SEARCH_TIME = 10000;
	
	/* Relative adjustment of average value of history thresholds and current thresholds */
	public static double HISTORY_ADJUSTMENT_OF_THRESHOLDS = 0.25;
	/* Relative increase of threshold profit */
	public static double PROFIT_THRESHOLD_INCREASE = 0.05;
	/* Relative decrease of thresholds if not successful*/
	public static double THRESHOLD_DECREASE = 0.1;
	
	/* Array of all current thresholds of the farme*/
	public double [] threshold = new double[NUMBER_OF_THRESHOLDS];
	
	//output history. The Output of DURATION_OF_HISTORY years are stored in this list. 
	//An output is a double[] with values of profit, labour, risk and legality
	public ArrayList <double[]> outputHistory;
	
	//average output history. The average output of the outputHistory
	public double [] averageOutputHistory = new double [NUMBER_OF_THRESHOLDS];
	
	//indicates if farmer has been just initialized
	boolean init = true;
	
	
	public Farmer_Satisficing(int type, int sizeClass) {
		super(type, sizeClass);
		outputHistory = new ArrayList<double[]>();
		setInitialOutputHistory();
		
	}
	
	

	/** This method adjusts all thresholds at the beginning of the year (threshold profit increase 
	 * 	 & history adjustment)
	 */
	public void increaseThresholds(){
		if (init){
			setInitialThresholds();
			init = false;
		}
		// first increase threshold profit
		if (profitIncluded){
			threshold [PROFIT] *= (1 + PROFIT_THRESHOLD_INCREASE);
		}
		
		// now adjust all thresholds (compare to averageOutputHistory)
		for (int i = 0; i < NUMBER_OF_THRESHOLDS; i++){
			threshold [i] +=  ((averageOutputHistory [i] - threshold [i] ) * HISTORY_ADJUSTMENT_OF_THRESHOLDS);
		}
		
	}
	
	/** 
	 * This method randomly chooses one threshold, that will be decreased
	 */
	public void reduceThreshold(){
		threshold[Random.uniform.nextIntFromTo(0, NUMBER_OF_THRESHOLDS-1)] *= (1-THRESHOLD_DECREASE);
	}
	
	/** This method updates the OutputHistory and updates the average Output Array
	 * @param currentOutput: The current Output
	 */
	public void updateOutputHistory (double [] currentOutput){
		
		//add current output
		outputHistory.add(currentOutput);
		
		//delete last if necessary
		if(outputHistory.size() > DURATION_OF_HISTORY){
			outputHistory.remove(0);
		}
		
		//update average Output
		updateAverageOutputHistory();
	}
	
	/** This Method updates the averageOutput - Array
	 * 
	 */
	public void updateAverageOutputHistory (){
		double[] average = new double [NUMBER_OF_THRESHOLDS];
		double[] tmp;

		ListIterator<double[]> it = outputHistory.listIterator();
		
		// all years in the list
		while (it.hasNext()){	
			tmp = it.next();
			
			//update all outputs
			for (int i = 0; i < NUMBER_OF_THRESHOLDS; i++){
				average[i] += (tmp[i]/outputHistory.size());
			}
			
			//overwrite averageOutputHistory with current average
			averageOutputHistory = average;
		}
	}
	
	/** This method sets the initial condition of the output History
	 * 
	 */
	public void setInitialOutputHistory(){
		//Is it needed?
		//It's not needed - the initial landUsePattern is the first Element in the ouputHistory
	}
	
	/** This method sets the initial Thresholds for the Farmer */
	public void setInitialThresholds(){
		
		/*for (int i=0; i < NUMBER_OF_THRESHOLDS; i++)
			threshold[i] = 0.5;
		if(profitIncluded)
			threshold[PROFIT] = 10000;*/
		//set initial thresholds on initial output
		for (int i=0; i< NUMBER_OF_THRESHOLDS; i++){
			threshold[i] = getCurrentOutput()[i];
			// If initial output is 0 at one scale (e.g. labour) set initial threshold = 0.1
			if (threshold[i]<0.1)
				threshold[i]=0.1;
//			if (i==LEGALITY){
//				threshold[i]=0.5;
//			}
		}
	}

	
	/** This method returns the thresholds of the farmer
	 *  @return thresholds of the farmer
	 */
	public double[] getThreshold(){
		return threshold;
	}
	
	/** This method returns the choice process of a Farmer_satisficing 
	 *  @overwrite
	 */
	protected static ChoiceProcess choiceProcessForType(int type){
		CollectOptions cO = null;
		if(GlobalVariables.DIFFUSION){
			cO = new CollectOptions_DiffusionNrPeople();
		}
		GenerateLandUsePattern gP = null;
		RuleApplication rA = null;
		// use compareLandUsePattern_Satisficing
		CompareLandUsePattern cP = new CompareLandUsePattern_Satisficing();
		
		//refinement will always be null as it doesn't take place
		ChoiceProcess cp = new ChoiceProcess(cO, gP, rA, cP, null, type);
		return cp;
	}
	
	/** This method returns the current output of all relevant thresholds */
	public double[] getCurrentOutput(){
		ChoiceProcess cp = this.choiceProcess;
		//this is possible - the Farmer_Satisficing's choice Process includes CompareLandUsePattern_Satisficing
		CompareLandUsePattern_Satisficing clps = (CompareLandUsePattern_Satisficing) cp.getComparePattern();
		double [] output = new double[NUMBER_OF_THRESHOLDS];
		// fill array
		if (profitIncluded)
			output [PROFIT] = clps.evaluateGrossMargin(elp, this);
		if (labourIncluded)
			output [LABOUR] = clps.evaluateLabour(elp, this);
		if (riskIncluded)
			output [RISK] = clps.evaluateRisk(elp, this);
		if (legalityIncluded)
			output [LEGALITY] = clps.evaluateLegality(elp, this);
		
		return output;
		
	}
	
	/**In each step a farmer reconsiders his landUsePattern as specified in f.choiceProcess.*/
	public void action(){

		//update OutputHistory
		updateOutputHistory(getCurrentOutput());
		
		// First increase Thresholds
		increaseThresholds();
		
		// get a new pattern
		EvaluatedLandUsePattern newPattern= choiceProcess.designLandPattern(this);
		
		//Clone pattern to conserve current attributes of Options
		setLandUsePattern(newPattern);
			
	}
}
