package model;

import java.util.ArrayList;
import java.util.Iterator;

/**A Farming object integrates a Crop and an IrrigationTechnology and provides functionality to
 * obtain information about such a combination regarding
 * - compatibility
 * - gross margin
 * - labour needs
 * - water needs
 * - uses of this combination in the community
 * @author gholtz
 *
 */
public class Farming implements Option {

	/**Refers to static types of Crop and IrrigationTechnology.
	 *Fields are organized as cropTechCompatibility[tech.type][crop.type]*/
	protected static final boolean[][] cropTechCompatibility = 
	//rainfed, flood, sprinkler,drip
	{{true,true,true, false},	//trad cereals
		{false,true,true,false},//high yield cereals
		{true,true,true, false},//sunflower
		{false,true,true, true},//sugarbeet
		{false,true,true, true}, //potatoes
		{true,true,false, true}, //vineyard
		{true,true,false, true}, //olives
		{false,true,false, true}, //melons
		{false,true,false, true}, //paprica
		{false,true,false, true},  //garlic
		{true, true, true, true}  //fallow
	};
	/**Static count of uses to provide global information. 
	 * Requires to set values in this array to 0 between model runs! */
	protected static int uses[][] = new int[11][4];
	
	/**Checks whether a Crop c can be irrigated using Technology t.
	 *Uses cropTechCompatibility[][] to obtain this information.*/
	public static boolean compatibility(Crop c, IrrigationTechnology t){
		
		return cropTechCompatibility[c.type][t.type];
	}
	
	/**This method searches "options" for a Farming object that combines cropType and 
	 * irrigationTechType. If no such Farming object is found, the method returns null.*/
	public static Option findFarmingOption(ArrayList<Option> options, int cropType, int irrigationTechType){
		
		Option o;
		Farming f;
		Iterator <Option> iterator = options.iterator();
		while(iterator.hasNext()){
			o = (Option) iterator.next();
			if(o instanceof Farming){
				f = (Farming) o;
				if(f.crop.type == cropType){
					if( ((f.tech == null)&&(irrigationTechType==-1)) || //should not happen
							(f.tech.type==irrigationTechType) ){
						return f;
					}
				}
			}
		}		
		return null;
	}	
	
	
	///////////////////////////////////////////////////////
	
	protected Crop crop;
	protected IrrigationTechnology tech;
	protected String name;
	
	
	public Farming(Crop crop, IrrigationTechnology tech){
		this.crop = crop;
		this.tech = tech;
	}
	
	public Farming(Crop crop, IrrigationTechnology tech, String name){
		this(crop, tech);
		this.name = name;
	}

	
	/** Gross Margin is calculated as income (price*yield) - costs (direct costs + 
	 * costs for water pumping). Costs for labour and subsidies are
	 * not included here. */
	public double grossMarginPerHa(double size){
		return incomePerHa(size) - costsPerHa(size);		
	}
	/**Returns this.crop.getPrice()*this.crop.yieldPerHa()*/
	public double incomePerHa(double size){
		
		return crop.getPrice()*yieldPerHa();  
	}
	/**Returns variable costs + costs for water (if applicable)*/
	public double costsPerHa(double size){
		
		double costs = crop.getVariableCosts(this.isIrrigated()) +
		waterUsePerHa()*GlobalVariables.waterPrice;
		return costs;
	}
	
	/*Labour is stored in Crop.*/
	public double labourPerHa(double size){

		if(!(this.isIrrigated())){
			//not irrigated
			return crop.labourNeedPerHa(size,false);
		}
		return crop.labourNeedPerHa(size,true);
	}
	/* Water needs of crop divided by water efficiency of technology.
	 * If no technology is used, water need is 0.0 (rainfed).
	 * Water use per ha is independent from size.*/
	public double waterUsePerHa(){
		
		if((tech != null)&&(tech.type != IrrigationTechnology.RAINFED)){
			return crop.getWaterNeed()/tech.waterEfficiency;
		}
		return 0.0;		//if no irrigation technology is used
						//-->rainfed farming --> no (ground-)water used		
	}
	
	public double yieldPerHa(){
		
		double yield;
		if((tech!= null)&&(tech.type != IrrigationTechnology.RAINFED)){
			yield = crop.getYieldIrrigated();
			
		}
		else{
			yield = crop.getYieldRainfed();
		}
		return yield;  
	}
	
	/**Refers to uses of Farming with same crop.type and tech.type*/
	public int getUses(){
		return uses[this.crop.type][this.tech.type];
	}
	public void setUses(int u){
		uses[this.crop.type][this.tech.type] = u;
	}
	
	public String getName(){
		if(name == null){
			if(crop.getType()==Crop.FALLOW){
				return crop.getName();
			}
			else{
				return crop.getName()+"_"+tech.getName();
			}
		}
		return name;
	}
	public void setName(String s){
		name = s;
	}
	
	/**Returns true if o is of type Farming and o.crop equals this.crop
	 * and o.tech equals this.tech*/
	public boolean equals(Option o){
		if(o instanceof Farming){
			Farming f = (Farming)o;
			if(crop.getType()==Crop.FALLOW){
				return f.crop.equals(crop);
			}
			else{
				return (f.crop.equals(this.crop)&& f.tech.equals(this.tech));
			}
		}
		//non-farming options are different from this Option
		return false;
	}
	
	/**Returns a Farming with same crop, cloned IrrigationTechnology and same name 
	 */
	public Farming clone(){
		Farming c = new Farming(this.crop, IrrigationTechnology.clone(this.tech));
		//Assumption: crop attributes remain similar or new varieties are
		//used by all farmers --> no need to clone crops. Problem with cloning
		//would be attribute price, which should change for all farmers.
		c.name = this.name;
		return c;
	}
	
	public Crop getCrop(){
		return crop;
	}
	public IrrigationTechnology getIrrigationTechnology(){
		return tech;
	}
	
	public boolean isIrrigated(){
		if( (this.tech != null)&&(this.tech.getType()!=IrrigationTechnology.RAINFED)){
			return true;
		}
		return false;
	}
	
	public String toString(){
		return getName();
	}
}
