package model;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;


/**A LandUsePattern combines several LandUses to represent the pattern of land use of
 * one farmer. The class implements the relevant features of a List, specified for LandUse and
 * provides aggregated information on gross margin, water used and labour needed.*/
public class LandUsePattern{
	
	/**Creates an exact copy of lp (clones each landUse).	 */
	public static LandUsePattern clone(LandUsePattern lp){
		
		LandUsePattern c = new LandUsePattern();
		for(int i=0; i<lp.length(); i++){
			c.add(LandUse.clone(lp.get(i)));
		}
		return c;
	}
	
	////////////////////////////////
	
	protected List<LandUse> landUses; 	
	
	public LandUsePattern(){
		landUses = new ArrayList<LandUse>(5);
	}
	/**Adds a LandUse. Does not check whether a LandUse with same Option
	 * is already included.
	 * @param l
	 * @return
	 */
	public boolean add(LandUse l){
		return landUses.add(l);
	}
	/**Length of list = number of different landUses.*/
	public int length(){
		return landUses.size();
	}
	
	public LandUse get(int index) throws IndexOutOfBoundsException{
		return landUses.get(index);
	}

	public LandUse remove(int index)throws IndexOutOfBoundsException{
		return landUses.remove(index);
	}
	
	/**Sum of sizes of landUses (= area).*/
	public double size(){
		double s = 0.0;
		LandUse l;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			s += l.size;
		}
		return s;
	}
	/**Returns int index for that get(index) returns a landUse l
	 * with l.option.equals(o). Returns -1 if no such LandUse could
	 * be found.	 */
	public int indexOf(Option o){
		
		for(int i=0; i< landUses.size(); i++){
			if(landUses.get(i).option.equals(o)){
				return i;
			}
		}
		return -1;
	}
	
	/***/
	public int indexOf(LandUse l){
		
		for(int i=0; i< landUses.size(); i++){
			if(landUses.get(i).equals(l)){
				return i;
			}
		}
		return -1;
	}
	
	/**Uses the equal(..) method of LandUse. Returns true if an equal landUse is included in this
	 * LandUsePattern.*/
	public boolean contains(LandUse l){
		for(int i=0; i< landUses.size(); i++){
			if(landUses.get(i).equals(l)){
				return true;
			}
		}
		return false;
		
	}
	
	/**Computes gross margin of this pattern: income()-costs().*/
	public double grossMargin(){

		double gm = income() - costs();
		/*		double gm = 0.0;
		LandUse l;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			gm += l.option.grossMarginPerHa(l.size)*l.size;
		}
*/
		return gm;
	}
	/** Sums up income of all land uses (uses option.incomePerHa(double s))*/
	public double income(){
		
		double income = 0.0;
		LandUse l;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			income += l.option.incomePerHa(l.size)*l.size;
		}
		return income;
	}
	/**Variable costs + costs for water (costs for labour are not(!) included*/
	public double costs(){
		
		double costs = 0.0;
		LandUse l;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			costs += l.option.costsPerHa(l.size)*l.size; //variable costs + costs for water 
		}

		return costs;
	}
	
	
	/**Sums up waterNeeds of all landUses combined in this pattern.*/
	public double waterUsed(){
		double water = 0.0;
		LandUse l;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			water += l.option.waterUsePerHa()*l.size;
		}
		return water;
	}
	
	/**Sums up labourNeeds of all landUses combined in this pattern.
	 * Total labour related to one crop is calculated as weighted average of irrigated 
	 * and non-irrigated crop areas.*/
	public double labour(){
		double labour = 0.0;
		
		double totalCropArea[] = new double[Crop.NR_CROPS];
		double irrigatedCropArea[] = new double[Crop.NR_CROPS];

		//sum up total and irrigated area of each crop
		LandUse l;
		Farming f;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			f = (Farming) l.option;
			totalCropArea[f.getCrop().getType()]+=l.size;
			if(f.isIrrigated()){
				irrigatedCropArea[f.getCrop().getType()]+=l.size;
			}
		}
		
		//for each crop calculate labour as weighted average of irrigated and rainfed labour needs and add to total labour
		double li=0.0;
		for(int i=0;i<totalCropArea.length; i++){
			if(irrigatedCropArea[i]<totalCropArea[i]){
				li=(Crop.labourNeedPerHa(i, totalCropArea[i],false)*(totalCropArea[i]-irrigatedCropArea[i])+
					Crop.labourNeedPerHa(i, totalCropArea[i],true)*(irrigatedCropArea[i]));
			}
			else{//only irrigated 
				li = Crop.labourNeedPerHa(i, totalCropArea[i],true)*totalCropArea[i];
			}
			labour += li;
		}
	
		return labour;
	}
	
	public String toString(){
		
		StringBuilder s = new StringBuilder(""); 
		LandUse l;
		for(int i=0; i< length(); i++){			
			l = get(i);
			s.append(l.getName()).append(": ").append(l.size).append(" ; ");
		}
		
		return s.toString();
		
	}	
	
	/**Returns area on which cropType is planted*/
	public double cropArea(int cropType){
		LandUse l;
		double a=0.0;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			if(((Farming)l.option).getCrop().getType()==cropType){
				a+=l.size;
			}
		}
		return a;
	}
	
	/**Returns area for which techType is used (works for IrrigationTechnology.Rainfed)*/
	public double irrigationTechArea(int techType){
		LandUse l;
		double a=0.0;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			if(((Farming)l.option).getIrrigationTechnology().getType()==techType){
				a+=l.size;
			}
		}
		return a;
	}
	/**Area irrigated (rainfed area is not considered)*/
	public double irrigatedArea(){
		LandUse l;
		double a=0.0;
		for(int i=0; i<landUses.size(); i++){
			l = landUses.get(i);
			if(((Farming)l.option).getIrrigationTechnology().getType()!=IrrigationTechnology.RAINFED){
				a+=l.size;
			}
		}
		return a; 
	}
	
	/**This method calculates the area of landUses in this landUsePattern
	 * specified by the parameter
	 * @param irrigated: if true only irrigated landUses are considered, if false only rainfed ones
	 * @param cropsConsidered: all landUses with cropsConsidered[landUse.crop.type]=true are considered
	 * @return
	 */
	public double areaOfLandUses(boolean irrigated, boolean[] cropsConsidered){
		
		double a = 0.0;
		LandUse l;
		Farming o;
		
		for(int i=0; i<this.length(); i++) {
			l = this.get(i);
			o = (Farming)l.option;
			if((o.isIrrigated()==irrigated) && cropsConsidered[o.crop.type]){
				a += l.size;
			}
		}
		return a;
	}
	
	public int nrCrops(){
		
		boolean [] cropsInLp = new boolean[model.Crop.NR_CROPS];
		Iterator<LandUse> it = landUses.iterator();
		
		while(it.hasNext()){
			cropsInLp[((model.Farming)(it.next().option)).getCrop().getType()]= true;
		}
		int nrCrops=0;
		for(int i=0;i<cropsInLp.length;i++){
			if(cropsInLp[i])nrCrops++;
		}
		return nrCrops;
	}

	public boolean equals(LandUsePattern lp){
		
		if(lp.length() != this.length()){
			return false;
		}
		boolean[] match = new boolean[this.length()];
		int index;
		for(int i=0; i< this.length(); i++){
			index = lp.indexOf(get(i));
			if(index <0){	//this LandUsePattern contains landUse (at index i) which is not contained in lp				
				return false;
			}
			else{
				match[index] = true;
			}
		}
		for(int i=0; i<match.length; i++){
			if(!match[i]){	//lp contains a landUse which is not contained in this LandUsePattern
				return false;
			}
		}
		return true;
	}
}
