package model.decisionMaking;

import model.Farmer_Satisficing;
import model.GlobalVariables;
import model.Option;
import model.rules.Rule;
import model.Farmer;

import java.util.ArrayList;

/**This class integrates the various sub-aspects of decision making (collecting options, generate landUsePattern,...) and
 * organises the program flow for the decision making of one Farmer.
 * This class draws on interfaces, therefore details of the decision making process can be changed
 * by using alternative classes without need to change this class.
 * 
 * This class has been edited by Marvin Nebel to include satisficing, changes are marked.
 * 
 * @author gholtz, (Marvin Nebel)
 *
 */


public class ChoiceProcess {
	
	protected CollectOptions collectOptions;
	protected  GenerateLandUsePattern generateLandUsePattern;
	protected  RuleApplication ruleApplication;
	protected  CompareLandUsePattern compareLandUsePattern;
	protected  RefinePattern refinePattern;
	
	/**Use null in fields to set to default classes which are
	 * - CollectOptions_allOptions()
	 * - GenerateLandUsePattern_localSearch()
	 * - RuleApplication_simple()
	 * - CompareLandUsePattern_WADD()
	 * - RefinePattern*/
	public ChoiceProcess(CollectOptions cO, GenerateLandUsePattern gP, RuleApplication rA,CompareLandUsePattern cP, RefinePattern rP, int farmer_type){
		
		//collectOptions
		if(cO != null){
			collectOptions =cO;
		}
		else{//default
			collectOptions = new CollectOptions_allOptions();
		}
		
		//generateLandUsePattern
		if(gP != null){
			generateLandUsePattern = gP;
		}
		else{
			generateLandUsePattern = new GenerateLandUsePattern_localSearch();
		}
		
		// ruleApplication
		if(rA!= null){
			ruleApplication = rA;
		}
		else{
			ruleApplication = new RuleApplication_simple();
		}
		
		
		//comparePattern
		if(cP != null){
			compareLandUsePattern = cP;
		}
		
		//CHANGE: use different compareLandUsePattern, if satisficing is used
		else{
			if (GlobalVariables.SATISFICING_PER_TYPE[farmer_type]){
				compareLandUsePattern = new CompareLandUsePattern_Satisficing();
			}
			else{
				compareLandUsePattern = new CompareLandUsePattern_Utility();
			}
		}
		
		//RefinePattern
		if(rP!=null){
			refinePattern = rP;
		}
		else{
			refinePattern = new RefinePattern();
		}		
	}
	
	/**This method has to be called to provide information to the respective
	 * sub-processes before designLandPattern(Farmer f) can be called.
	 * options, farmers and rules must not be null.  */
	//CHANGE: This method has been changed for reasons in design in the satisficing model. 
	// The new argument int number_of_farmers replaces the former argument ArrayList <farmer> farmer
	public void setUp(ArrayList<Option> options, ArrayList<Rule>rules, int number_of_farmers){
		
		collectOptions.setAllOptions(options);
		ruleApplication.setRules(rules);
		if(collectOptions instanceof CollectOptions_DiffusionNrPeople){
			((CollectOptions_DiffusionNrPeople) collectOptions).setFullUsage(number_of_farmers);
		}	
	}
	
	/**Returns the "best" new landUsePattern for f by applying the different
	steps collect options, generate pattern, evaluate pattern, choosePattern,
	refinePattern,... . Uses f.landUsePattern as starting point.*/
	public EvaluatedLandUsePattern designLandPattern(Farmer f){
		
		EvaluatedLandUsePattern bestPattern;
		
		//CHANGE: A new step has been included for the satisficing model.
		//step 0: if satisficing is included: check current pattern first (for reasons of performance)
		if (GlobalVariables.SATISFICING_PER_TYPE[f.type]){
			//cast compareLandUsePattern to compareLandUsePattern_Satisficing and Farmer to Farmer_Satisficing
			CompareLandUsePattern_Satisficing cps = (CompareLandUsePattern_Satisficing) compareLandUsePattern;
			Farmer_Satisficing fs = (Farmer_Satisficing) f;
			//create ArrayList
			ArrayList<EvaluatedLandUsePattern> current = new ArrayList <EvaluatedLandUsePattern>();
			// add current elp as the only elp in the list
			current.add(fs.elp);
			//apply rules to current elp
			current = ruleApplication.applyRulesTo(current, fs);
			
			//check whether elp is satisficing
			if (cps.isSatisficing(current.get(0), fs))
			{	
				return current.get(0);
			}
		}
		
		//step 1: collect all Options considered by this farmer
		Option [] consideredOptions = collectOptions.collectOptionsForFarmer(f);


		//step 2: Creating LandUsePattern
		ArrayList<EvaluatedLandUsePattern> elpList = generateLandUsePattern.createPattern(consideredOptions, f);
		
		//step 3: filter pattern with regard to distance 
		elpList = compareLandUsePattern.discardPattern(elpList, f);

		//step 4: evaluate residual pattern
		elpList = ruleApplication.applyRulesTo(elpList, f);

		//step 4a: remove all patterns that return 0 at the legality output (ONLY FOR SATISFICING)
		if (GlobalVariables.SATISFICING_PER_TYPE[f.type]){
			//cast compareLandUsePattern to compareLandUsePattern_Satisficing and Farmer to Farmer_Satisficing
			CompareLandUsePattern_Satisficing cps = (CompareLandUsePattern_Satisficing) compareLandUsePattern;
			elpList = cps.discardPattern_Legality(elpList, f);
		}
//*debug
		if(elpList.size() <= 0){
			System.err.println("ChoiceProcess: elpList is empty!");	
		}//*/		

		//step 5: compare pattern and choose one
		bestPattern = compareLandUsePattern.choosePattern(elpList,f);

		//CHANGE: include step 5a for satisficing
		//step 5a: if Satisficing: if no pattern was found, reduce threshold and try again until a pattern is found
		int searchtime = 0;
		if (compareLandUsePattern instanceof CompareLandUsePattern_Satisficing){
			while (bestPattern == null){
				searchtime++;
				// check if farmer is of instance Farmer_Satisficing
				if (!(f instanceof Farmer_Satisficing)){
					/*debug*/  System.err.println("ChoiceProcess: Using CompareLandUsePattern_Satisficing but not Farmer_Satisficing class!");
					return null;
				}
				// cast Farmer to satisficing type
				Farmer_Satisficing fs = (Farmer_Satisficing)f;
				fs.reduceThreshold();
				bestPattern = compareLandUsePattern.choosePattern(elpList,fs);
				
				/* Debug*/ //if no satisficing pattern can be found after MAX_SEARCH_TIME
				if (searchtime > Farmer_Satisficing.MAX_SEARCH_TIME && bestPattern == null){
					System.err.println("ChoiceProcess: " +
							"No satisfactory pattern could be found after " + searchtime + "rounds! " +
									"The current pattern is chosen.");
					//create ArrayList
					ArrayList<EvaluatedLandUsePattern> current = new ArrayList <EvaluatedLandUsePattern>();
					// add current elp as the only elp in the list
					current.add(fs.elp);
					//apply rules to current elp
					current = ruleApplication.applyRulesTo(current, fs);
					bestPattern = current.get(0);
				}
			}
		}
		
		//CHANGE: Step 6 only within utility-version
		//step 6: refine this best pattern only for CompareLandUsePattern_Utility
		if (compareLandUsePattern instanceof CompareLandUsePattern_Utility){
			elpList.clear();
			//step 6a: create set of refined pattern
			elpList = refinePattern.refine(bestPattern,f);
			//step 6b: evaluate set of refined pattern
			ruleApplication.applyRulesTo(elpList, f);
			//step 6c: choose best one
			bestPattern = compareLandUsePattern.choosePattern(elpList,f);
		}
		
		return bestPattern;
		
	}
	
	public CompareLandUsePattern getComparePattern(){
		return compareLandUsePattern;
	}
}
