/**
 * 
 */
package model.decisionMaking;

import java.util.ArrayList;
import java.util.Iterator;

import uchicago.src.sim.util.Random;

import model.Distances;
import model.Farmer;
import model.Farmer_Satisficing;
import model.GlobalVariables;
import model.LandUsePattern;

/** This class is to choose a satisficing Land Use Pattern for a Farmer f
 * @author Marvin Nebel
 * @version 03/29/2011
 */
public class CompareLandUsePattern_Satisficing implements CompareLandUsePattern {

	public static double [] LABOUR_LIMITS = {0.6,2.0,5.0};
	public static double [] SKILLS_SHAPES = {10.0,6.0,4.0};
	public static double [] CAPITAL_SHAPES = {4.0,6.0,4.0};
	/**
	 * Chooses the first evaluated landuse pattern in elpList, that is satisfactory for f
	 * 
	 */
	public EvaluatedLandUsePattern choosePattern(
			ArrayList<EvaluatedLandUsePattern> elpList, Farmer f) {
		
		if((elpList == null) || (elpList.isEmpty())){ 
			/*debug*/  System.err.println("CompareLandUsePattern_Satisficing.choosePattern: empty list!");
						return null;
					}
		
	
		// check if farmer is of instance Farmer_Satisficing
		if (!(f instanceof Farmer_Satisficing)){
			/*debug*/  System.err.println("CompareLandUsePattern_Satisficing.choosePattern: using wrong farmer class");
			return null;
		}
		// cast Farmer to satisficing type
		Farmer_Satisficing fs = (Farmer_Satisficing)f;
		
		
		Iterator<EvaluatedLandUsePattern> it = elpList.iterator();
		EvaluatedLandUsePattern elp;
		
		//check new considered landuse-pattern as long as no satisficing pattern is found
		while(it.hasNext()){	
			elp = it.next();
			
			if(elp != null){
				// now check, if elp is satisfactory
				if (isSatisficing(elp, fs)){
					// if pattern is satisfactory return the chosen pattern
					return elp;
				}
				
			}
		}
		
		//if no pattern was found return null
		return null;
	}

	
	
	
	/**
	 * Discards evaluated landuse pattern elp with probability depending on distance d of 
	 * elp to f. Probability to keep a pattern is p = Math.pow(1.0-d,GlobalVariables.distanceShape)
	 * The first element in elpList is definitively kept. (see CompareLandUsePattern_Utility)
	 * CHANGE in the satisficing model: 
	 * a Pattern that includes at least one option, that captures less than 10% or less than 1ha of the total size
	 *  will be deleted!! 
	 */
	public ArrayList<EvaluatedLandUsePattern> discardPattern(
			ArrayList<EvaluatedLandUsePattern> elpList, Farmer f) {
		
		if((elpList == null) || (elpList.isEmpty())){
			return elpList;
		}
		//collect here all elp that should be deleted (directly deleting
		//it from elpList while iterating over elpList leads to ConcurrentModificationException)
		ArrayList<EvaluatedLandUsePattern> delete = new ArrayList<EvaluatedLandUsePattern>();
		
		Iterator<EvaluatedLandUsePattern> it = elpList.iterator();
		
		it.next();//keep first entry in list = f's old landuse pattern		
	
		EvaluatedLandUsePattern elp;
		
		//Discard all that are too complex (more than maxNrCrops different crops 
		//or more than maxNrLU different land-uses*/
		int maxNrCrops =5;
		int maxNrLU = 7;
		
		while(it.hasNext()){
			elp = it.next();
			
			if((elp!=null) && ((elp.pattern.length() > maxNrLU) || (elp.pattern.nrCrops()>maxNrCrops))){
				delete.add(elp);
			}
			
			// CHANGE: delete all LandUses with options that capture less than 10% and 1ha of total size	
			if (GlobalVariables.DISCARDSMALLPATTERN){
				for(int i=0; i<elp.pattern.length();i++){
					double s = elp.pattern.get(i).size;
					if((s < f.sizeOfFarm*0.1) && (s < 1.0))
						delete.add(elp);
				}
			}
			// CHANGE: delete all LandUses that return
			// an output of profit <= 0, labour <= 0 or risk <= 0 for performance reasons
			if((evaluateGrossMargin(elp,f) <= 0 ) || (evaluateLabour(elp, f) <= 0) || evaluateRisk(elp, f)<=0)
				delete.add(elp);
		
		}
		


		//Discard according to skills
		it = elpList.iterator();
		it.next(); //skip first element = last years lp
		while(it.hasNext()){	//check all considered landuse-pattern
			elp = it.next();
			if(elp != null){
				//Distance als Filter
				double d = Distances.getDistanceSkills(elp.pattern, f);		
				
				//compute distance
				double p = Math.pow(1.0-d,SKILLS_SHAPES[f.type]);

				//compute probability that elp.pattern will be considered
				if(Random.uniform.nextDouble()>=p){							//elp should be kept
					delete.add(elp);
				}
			}
		}
		
		
		//Discard according to capital and investments
//		ArrayList<EvaluatedLandUsePattern> keepPattern = new ArrayList<EvaluatedLandUsePattern>();
//		it = keepPatternSkills.iterator();	//consider only those "close" enough according to skills

		it = elpList.iterator();
		it.next();
//		keepPattern.add(it.next());						//keep first element (=last years land use)
		while(it.hasNext()){	//check all considered landuse-pattern
			elp = it.next();
			if(elp != null){
				//Distance als Filter
				double d = Distances.getDistanceCapital(elp.pattern, f);			//compute distance
				double p = Math.pow(1.0-d,CAPITAL_SHAPES[f.type]);	//compute probability that elp.pattern will be considered
				if(Random.uniform.nextDouble()>=p){							//elp should be discarded
					delete.add(elp);
				}
			}
		}
		elpList.removeAll(delete);
		return elpList;
	}
	
	
	
	/** This method checks, weather a LandUse is satisfactory for a Farmer or not
	 *  Are all thresholds met or not?
	 */
	public boolean isSatisficing (EvaluatedLandUsePattern elp, Farmer_Satisficing f){
		// get output of elp
		double profit = evaluateGrossMargin(elp, f);
		double labour = evaluateLabour(elp, f);
		double risk = evaluateRisk(elp, f);
		double legality = evaluateLegality(elp, f);
		
		// Is threshold profit included?
		if (Farmer_Satisficing.profitIncluded){
			// Is threshold reached?
			if (profit < f.threshold[Farmer_Satisficing.PROFIT]){
				return false;
			}
		}
			
			// Is threshold labour included?
			if (Farmer_Satisficing.labourIncluded){
				// Is threshold reached?
				if (labour < f.threshold[Farmer_Satisficing.LABOUR]){
					return false;
				}
			}
			
			// Is threshold risk included?
			if (Farmer_Satisficing.riskIncluded){
				// Is threshold reached?
				if (risk < f.threshold[Farmer_Satisficing.RISK]){
					return false;
				}
			}
			
			// Is threshold legality included?
			if (Farmer_Satisficing.legalityIncluded){
				// Is threshold reached?
				if (legality < f.threshold[Farmer_Satisficing.LEGALITY]){
					return false;
				}
			}
		// All thresholds are met
		return true;
	}
	
	
	// The following three methods are adopted from CompareLandUsePattern_Utility
	
	public double evaluateGrossMargin(EvaluatedLandUsePattern elp, Farmer f){
		LandUsePattern lp = elp.pattern;
		double gm = (lp.grossMargin() - f.labourCosts(lp) +elp.subsidies()-elp.penalties());
		
		return gm;
	}
	
	public double evaluateRisk(EvaluatedLandUsePattern elp, Farmer f){
		return 1.0-RiskEvaluation.riskFor(f,elp);
	}
	
	public double evaluateLabour(EvaluatedLandUsePattern elp, Farmer f){
		
		//maximum labour load 
		double k = LABOUR_LIMITS[f.type]; 
		double l = elp.pattern.labour(); 
		double fl = f.familyLabour;
		if(l>=k){				//amount of labour for elp exceeds capacity of farm
			return 0.0;
		}
		else if(l<=fl){			//amount of labour can be handled by family labour force
			return 1.0;
		}
		else{
			return 1.0-(l-fl)/(k-fl);
		}
	}
	
	
	// This method has been changed for the satisficing model.
	// (see EvaluatedLandUsePattern)
	public double evaluateLegality(EvaluatedLandUsePattern elp, Farmer f){
		return elp.legality();
	}
	
	


	/**This Method discards all Pattern that return a legality-output of 0
	 * 
	 * @param elpList
	 * @param f
	 * @return
	 */
	public ArrayList<EvaluatedLandUsePattern> discardPattern_Legality(
			ArrayList<EvaluatedLandUsePattern> elpList, Farmer f) {
		if((elpList == null) || (elpList.isEmpty())){
			return elpList;
		}
		//collect here all elp that should be deleted (directly deleting
		//it from elpList while iterating over elpList leads to ConcurrentModificationException)
		ArrayList<EvaluatedLandUsePattern> delete = new ArrayList<EvaluatedLandUsePattern>();
		
		Iterator<EvaluatedLandUsePattern> it = elpList.iterator();
		
		EvaluatedLandUsePattern elp;
		
		while(it.hasNext()){
			elp = it.next();
			if (evaluateLegality(elp, f)<=0){
				delete.add(elp);
			}			
		}
		elpList.removeAll(delete);
		
		return elpList;
	}

}


