package model.decisionMaking;

import uchicago.src.sim.util.Random;
import model.Distances;
import model.Farmer;
import model.LandUsePattern;

import java.util.ArrayList;
import java.util.Iterator;

/**Choose best LandUsePattern 
 * 
 * @author Georg
 *
 */
public class CompareLandUsePattern_Utility implements CompareLandUsePattern {
	
	public static double [] GM_SHAPES = {0.5,0.5,1.0};
	public static double [] LABOUR_SHAPES = {1.0,1.0,1.0};
	public static double [] LABOUR_LIMITS = {0.6,2.0,5.0};
	public static double [] RISK_SHAPES = {2.0,5.0,5.0};
	public static double [] LEGAL_SHAPES = {3.0,0.5,3.0};

	public static double [] SKILLS_SHAPES = {50.0,30.0,20.0};
	public static double [] CAPITAL_SHAPES = {20.0,30.0,20.0};
	
	
	/**
	 * Discards evaluated landuse pattern elp with probability depending on distance d of 
	 * elp to f. Probability to keep a pattern is p = Math.pow(1.0-d,GlobalVariables.distanceShape)
	 * The first element in elpList is definitively kept.
	 */
	public ArrayList<EvaluatedLandUsePattern> discardPattern(ArrayList<EvaluatedLandUsePattern> elpList, model.Farmer f){
		
		if((elpList == null) || (elpList.isEmpty())){
			return elpList;
		}
		//collect here all elp that should be deleted (directly deleting
		//it from elpList while iterating over elpList leads to ConcurrentModificationException)
		ArrayList<EvaluatedLandUsePattern> delete = new ArrayList<EvaluatedLandUsePattern>();
		
		Iterator<EvaluatedLandUsePattern> it = elpList.iterator();
		
		it.next();//keep first entry in list = f's old landuse pattern		
	
		EvaluatedLandUsePattern elp;
		
		//Discard all that are too complex (more than maxNrCrops different crops 
		//or more than maxNrLU different land-uses*/
		int maxNrCrops =5;
		int maxNrLU = 7;
		
		while(it.hasNext()){
			elp = it.next();
			
			if((elp!=null) && ((elp.pattern.length() > maxNrLU) || (elp.pattern.nrCrops()>maxNrCrops))){
				delete.add(elp);
			}
		}
		
		//Discard according to skills
		it = elpList.iterator();
		it.next(); //skip first element = last years lp
		while(it.hasNext()){	//check all considered landuse-pattern
			elp = it.next();
			if(elp != null){
				//Distance als Filter
				double d = Distances.getDistanceSkills(elp.pattern, f);			//compute distance
				double p = Math.pow(1.0-d,SKILLS_SHAPES[f.type]);	//compute probability that elp.pattern will be considered
				if(Random.uniform.nextDouble()>=p){							//elp should be kept
					delete.add(elp);
				}
			}
		}
		
		
		//Discard according to capital and investments
//		ArrayList<EvaluatedLandUsePattern> keepPattern = new ArrayList<EvaluatedLandUsePattern>();
//		it = keepPatternSkills.iterator();	//consider only those "close" enough according to skills

		it = elpList.iterator();
		it.next();
//		keepPattern.add(it.next());						//keep first element (=last years land use)
		while(it.hasNext()){	//check all considered landuse-pattern
			elp = it.next();
			if(elp != null){
				//Distance als Filter
				double d = Distances.getDistanceCapital(elp.pattern, f);			//compute distance
				double p = Math.pow(1.0-d,CAPITAL_SHAPES[f.type]);	//compute probability that elp.pattern will be considered
				if(Random.uniform.nextDouble()>=p){							//elp should be discarded
					delete.add(elp);
				}
			}
		}
		elpList.removeAll(delete);
		return elpList;
	}
	
	/**
	 * Chooses the best evaluated landuse pattern in elpList for f, according to 
	 * a utility as defined by this.evaluatePattern(elp, f).
	 */
	public EvaluatedLandUsePattern choosePattern(ArrayList<EvaluatedLandUsePattern> elpList,Farmer f){
		
		if((elpList == null) || (elpList.isEmpty())){ 
/*debug*/  System.err.println("CompareLandUsePattern.choosePattern: empty list!");
			return null;
		}
		double utility=0.0;
		double bestUtility = -Double.MAX_VALUE;	
		EvaluatedLandUsePattern best = null;
		
		Iterator<EvaluatedLandUsePattern> it = elpList.iterator();
		EvaluatedLandUsePattern elp;
		while(it.hasNext()){	//check all considered landuse-pattern
			elp = it.next();
			if(elp != null){
				
				utility = this.evaluatePattern(elp, f);					//compute utility and choose if better
					
				if(utility>bestUtility){
					bestUtility = utility;
					best = elp;
				}
				else if((utility ==bestUtility)&& (Random.uniform.nextDouble()<0.5)){
					bestUtility = utility;
					best = elp;

				}
			}		
		}
		//the following should never be needed, since utility 
		//of previous landuse pattern is > -Double.MAX_VALUE
		if(best == null){ //if no pattern was chosen
			it = elpList.iterator();	//use previous pattern (first in list)
			best = it.next();
		}
		return best;
	}
	
/**
 * Computes utility of elp for f using gross margin, this.evaluateRisk(elp,f), 
 * this.evaluateLabour(elp,f), this.evaluateLegality(elp,f). Further 
 * f.priorityFactors[] and GlobalVariables.XYFactor are used.
 */
	public double evaluatePattern(EvaluatedLandUsePattern elp, Farmer f){
		double utility;
		
		double gm = evaluateGrossMargin(elp,f);
		if(gm == 0.0){
			utility = 0.0;
		}
		else{
			utility = Math.signum(gm)*Math.pow(Math.abs(gm),GM_SHAPES[f.type]);
		}
		
		utility = utility * Math.pow(evaluateRisk(elp,f),RISK_SHAPES[f.type]);

		utility = utility * Math.pow(evaluateLabour(elp,f),LABOUR_SHAPES[f.type]);
		
		utility = utility * Math.pow(evaluateLegality(elp,f),LEGAL_SHAPES[f.type]);

		return utility;
	}
	
	public double evaluateGrossMargin(EvaluatedLandUsePattern elp, Farmer f){
		LandUsePattern lp = elp.pattern;
		double gm = (lp.grossMargin() - f.labourCosts(lp) +elp.subsidies()-elp.penalties());
		
		return gm;
	}
	
	
	
	public double evaluateRisk(EvaluatedLandUsePattern elp, Farmer f){
		return 1.0-RiskEvaluation.riskFor(f,elp);
	}
	
	
	
	public double evaluateLegality(EvaluatedLandUsePattern elp, Farmer f){
		//if complies with both laws, return 0.0, if at least one law
		//is violated, return 1.0
		return elp.violatesLaw() ? 0.5:1.0;
	}
	
	
	public double evaluateLabour(EvaluatedLandUsePattern elp, Farmer f){
		
		//maximum labour load 
		double k = LABOUR_LIMITS[f.type]; 
		double l = elp.pattern.labour(); 
		double fl = f.familyLabour;
		if(l>=k){				//amount of labour for elp exceeds capacity of farm
			return 0.0;
		}
		else if(l<=fl){			//amount of labour can be handled by family labour force
			return 1.0;
		}
		else{
			return 1.0-(l-fl)/(k-fl);
		}
	}
}
