package model.decisionMaking;

/**This class generates landUsePattern based on a starting pattern
 * by applying combinations of the following procedures:
 * - rescaling the areas of two Options
 * - adding one Option
 * - dropping one Option. 
 * 
 * It also considers all specializations on single Options.
 * @author Georg
 *
 */

import uchicago.src.sim.util.Random;
import model.LandUsePattern;
import model.Option;
import model.Farmer;
import model.LandUse;

import java.util.ArrayList;

public class GenerateLandUsePattern_localSearch implements GenerateLandUsePattern{
	
	public int nrPatternToCreate=1000;	//defines how many pattern are considered and evaluated within one time step by one farmer
	public int maxCombinedChanges = 20;	//defines how different a pattern can get within one time step (how often a change method is called
	public double pCombine = 0.8;		//after one change method has been called, probability to add further changes to one pattern 
	
	//probs for different change methods. Probability of "Rescale" is the remainder (1-pAdd-pDrop)
	public double pAdd = 0.2;
	public double pDrop = 0.2;
	public double pReplace = 0.2;
	
	//If true, specializations to one crop only are always included as 
	//LandUsePattern in the array created in createPattern(..); additional to
	//the nrPatternToCreate  
	public boolean addSpecialists = true;
	
	
	/**Creates pattern and includes current landUsePattern of f as first entry in return.
	 * array.	 */
	public ArrayList<EvaluatedLandUsePattern> createPattern(Option[] o, Farmer f){
		

		int nrPatternGenerate = nrPatternToCreate; 	
		int nrPatternTotal = nrPatternGenerate+1;	////add space for current pattern
		if(addSpecialists) nrPatternTotal += o.length; //add space for specializations 
		
		ArrayList<EvaluatedLandUsePattern> p = new ArrayList<EvaluatedLandUsePattern>();

		p.add(new EvaluatedLandUsePattern(f.landUsePattern()));	//add farmer's current LandUsePattern
		
		for(int i=0; i<nrPatternGenerate; i++){	//create new pattern based on current pattern
			p.add(new EvaluatedLandUsePattern(createNextPattern(o,f)));
		}
		
		if(addSpecialists){	//add specialisations 
			LandUsePattern lp;
			LandUse l;
			for(int i=0; i< o.length; i++){
				int index = f.landUsePattern().indexOf(o[i]);	//consider here only specialization to landUses a Farmer is familiar with
				if( index >=0){
					lp = new LandUsePattern();				
					l = LandUse.clone(f.landUsePattern().get(index));
					l.size = f.landUsePattern().size();
					
					lp.add(l);
					p.add(new EvaluatedLandUsePattern(lp));
				}
/* this section can be added to make all kind of 
//specialisations directly possible, not only those already known to
// the farmer				 	
				else{
					lp = new LandUsePattern();
					l = new LandUse();
					l.option = o[i].clone();
					l.size = f.landUsePattern().size();
					
					lp.add(l);
					p.add(new EvaluatedLandUsePattern(lp));
				}
//*/
			}
		}
		return p;
	}

	
	/**Creates a new LandUsePattern lp based on f.landUsePattern and using the Options
	 * in o. lp is created by applying change methods (rescale, add, drop, replace) to
	 * f.landUsePattern.
	 * Assumption: more radical changes are less frequent, therefore, combinations
	 * of changes become the less likely, the more steps are combined.	 */
	protected LandUsePattern createNextPattern(Option[] o, Farmer f){
		
//*debug*/if((o == null) || (o.length == 0))System.err.println("empty option array:"+f+" landUse:"+f.landUsePattern());
		LandUsePattern current = f.landUsePattern();
		
		//clone current landUsePattern, then change it
		LandUsePattern l = LandUsePattern.clone(current);
		
		double diceCombinations; 
		int step = 0;	
		//do at least one change step, probably more, up to maxCombinedChanges
		do{	
			//what type of change (add, drop, replace, rescale)?
			double diceChangeType = Random.uniform.nextDouble();
			if(diceChangeType < pAdd){
				addOption(l,o,f);
			}
			else{
				if(diceChangeType < pAdd+pDrop){
					dropOption(l);	
				}
				else{
					if(diceChangeType < pAdd+pDrop+pReplace){
						replaceOption(l,o);
					}
					else{
						rescale(l);
					}
				}	
			}
			
			step++;
			diceCombinations = Random.uniform.nextDouble();
			//diceCombinations makes more radical less likely, maxCombinedChanges sets upper limit
		}while((diceCombinations < pCombine) && (step < maxCombinedChanges));
	
		return l;
	}
	/**Randomly chooses one landUse which is dropped (if not only one)
	and replaces free area through expansion of another (random) landUse.*/
	protected boolean dropOption(LandUsePattern l){
		
		if(l.length() <=1){ //do not drop if only one landUse
			return false;
		}
		//randomly choose one landUse which is dropped
		int toDrop = (int) (Random.uniform.nextDouble()*l.length());
		//randomly choose another landUse which will be expanded
		int expand = (toDrop + 1 + ((int) (Random.uniform.nextDouble()*l.length()-1))) % l.length();
		double exchangedArea = l.get(toDrop).size;
		l.get(expand).size += exchangedArea;
		l.remove(toDrop);
		return true;
	}
	/**New Option is introduced on random area (<10% of farmSize). Area is taken from
	 * random current landUse l2. If area of new option is bigger than that of 
	 * l2, remove l2 (area exchanged = l2.size) */
	protected boolean addOption(LandUsePattern l, Option[] o, Farmer f){	
		//Randomly choose Option not yet part of l
		int oIndex = (int) (Random.uniform.nextDouble()*o.length);
		Option newO = null;
		int index;
		for(int i=0; i<o.length; i++){
			index = (oIndex+i)%o.length;
			if(l.indexOf(o[index])<0){	//l does not contain o[(oIndex+i)%o.length]
				newO = o[index];
				break;
			}
		}
		if(newO == null){	//no new Option found
			return false;
		}
		
		//randomly choose a landUse which will be reduced
		int indexReduce = (int) (Random.uniform.nextDouble()*l.length());	
		LandUse reduce = l.get(indexReduce);
		//Area exchanged is min of random size (<50% farmSize) and size of current landUse
//		double exchangedArea = Math.min(Random.uniform.nextDouble()*f.sizeOfFarm*0.5,reduce.size);
		double exchangedArea = Random.uniform.nextDouble()*reduce.size;
		LandUse newUse = new LandUse();
		newUse.option = newO.clone();
		newUse.size = exchangedArea;
		l.add(newUse);
		//if exchangedArea is complete area of current Use, remove current Use
		if(exchangedArea >= reduce.size){
			l.remove(indexReduce);
		}
		else{
			reduce.size -= exchangedArea;
		}
		return true;	
	}
	/**Replace one option with option not used yet.*/
	protected boolean replaceOption(LandUsePattern l, Option[] o){
		
		int oIndex = (int) (Random.uniform.nextDouble()*o.length);
		Option newO = null;
		int index;
		for(int i=0; i<o.length; i++){
			index = (oIndex+i)%o.length;
			if(l.indexOf(o[index])<0){	//l does not contain o[(oIndex+i)%o.length]
				newO = o[index];
				break;
			}
		}
		if(newO == null){	//no new Option found
			return false;
		}
		//randomly choose a landUse which will be replaced
		int indexReplace = (int) (Random.uniform.nextDouble()*l.length());	
		LandUse replace = l.get(indexReplace);
		LandUse newUse = new LandUse();
		newUse.option = newO.clone();
		newUse.size = replace.size;
		l.remove(indexReplace);
		l.add(newUse);
		return true;
	}
	
	
	/**Re-scales area assigned to two landUses. LandUses are chosen randomly. 
	 * Area rescaled is randomly chosen in [0,min(l1.size, l2.size)].
	 */
	protected boolean rescale(LandUsePattern l){
		
		if(l.length() <2){ //only one landUse --> nothing can be rescaled
			return false;
		}
		
		int indexL1 = (int) (Random.uniform.nextDouble()*l.length());
		int indexL2 = (indexL1 + ((int) (Random.uniform.nextDouble()*l.length()-1))) % l.length();
		
		LandUse l1 = l.get(indexL1);
		LandUse l2 = l.get(indexL2);
		
		double exchangedArea = Random.uniform.nextDouble()*Math.min(l1.size,l2.size);
		
		//exchange area: increase l1, reduce l2
		l1.size += exchangedArea;
		
		if(exchangedArea >= l2.size){ //only for case that random yields 0.0
			l.remove(indexL2);
		}
		else{
			l2.size -= exchangedArea;
		}
		
		return true;
	}
}
