package model.rules;

import java.util.HashMap;

import model.*;
import model.decisionMaking.EvaluatedLandUsePattern;

public class AEP extends Rule {

	public static int NO_AEP = 0;
	public static int OPTION100 = 3;
	public static int OPTION70 = 2;
	public static int OPTION50 = 1;
	
	protected PumpingQuotas pq;
	
	protected HashMap<Farmer,Double> aepArea;
	
	public AEP(PumpingQuotas pq){
		this.pq = pq;
	}
	
	public void prepareNextStep(){	
		if(getYear() == 1993){
			yc.setXLabel("AEP");
		}
		if(getYear() == 2003){
			yc.setXLabel("MOD");
		}
	}
	
	
	public void evaluateLandUsePattern(EvaluatedLandUsePattern elp, Farmer f) {
			
		
		//default: no AEP
		elp.aepOption = AEP.NO_AEP;
		elp.aepCompPayments = 0.0;
		elp.aepArea = 0.0;
	
		double rights = waterRights(f,elp,getYear());	
		
		//before 1993 or no water rights-> no possibility to participate in AEP
		if(rights>=0){
			double percentWater = elp.pattern.waterUsed() /rights * 100;
			if(percentWater <= 0.0) elp.aepOption = AEP.OPTION100;
			else if((percentWater<=30.0) && (getYear()<2003)) elp.aepOption = AEP.OPTION70;//70% option only available before 2003
			else if(percentWater<= 50.0) elp.aepOption = AEP.OPTION50;
			
			
			if(elp.aepOption != AEP.NO_AEP){
	
				elp.aepArea = aepArea(f,elp,getYear());
				elp.aepCompPayments = payments(getYear(),elp.aepOption,elp.aepArea);
				elp.complyWithWaterLaw = true;
			}	
		}
	}
	
	
	/**Returns water rights of Farmer f. If f is null or no water rights are memorized, the
	 * return value is -1.0.
	 */

	
	/**option must be NO_AEP, OPTION50, OPTION70 or OPTION100.*/
	public double payments(int year, int option, double area){
		
		//	period in [0,4], [1985-1992]->0, [1993-1996]->1, [1997-2000]->2, [2001-2002]->3,[2003...]->4
		int period;
		
		/*payments for periods 0..3 and for options NO_AEP, OPTION50, OPTION70, OPTION100*/
		double p03[][] = {{0.0,0.0,0.0,0.0},{0.0,156.3,258.4,360.6},
				{0.0,164.1,271.4,378.6},{0.0,179.4,296.6,414.1}};
		/*payments for period 4, sizes
		 * <=40ha, 40-80ha, >80ha and options NO_AEP, OPTION50, OPTION70, OPTION100 
		 * Option70 is a dummy, it does not exist for that period*/
		double p4[][] = {{0.0, 209.0, 209.0,518.0}, {0.0, 125.0,125.0,311.0}, {0.0, 63.0,63.0, 155.0}};
//		double p4[][] = {{0.0, 0.0, 0.0,0.0}, {0.0, 0.0,0.0,0.0}, {0.0, 0.0,0.0, 0.0}};
		
		if(year < 1993){
			period = 0;
		}
		else if(year < 1997){ 
				period = 1;
			}
		else if(year < 2001){
			period = 2;
		}
		else if(year < 2003){
			period = 3;
		}
		else{
			period = 4;
		}
		
		if(period <4){
			return p03[period][option]*area;
		}
		else{ //modulation
			/* int s; */
			double pay =0;
			double a = area;
			//first 40ha
			pay+=p4[0][option]*Math.min(a,40.0);
			a = a-40.0;
			//40-80ha
			if(a>0.0){
				pay+=p4[1][option]*Math.min(a,40.0);
				a = a-40.0;
			}
			////rest (>80ha)
			if(a>0.0){
				pay+=p4[2][option]*a;
			}
			return pay;
	/*		if(area<=40){
				s = 0;
			}
			else if(area <=80){
				s = 1;
			}
			else{
				s = 2;
			}
			return p4[s][option];*/
		}
	}
	
	public double waterRights(Farmer f, EvaluatedLandUsePattern elp, int year){
		
		//AEP not yet introduced
		if(getYear()<=1992){
			return 0.0;
		}
//		first phase
		if(getYear()<2003){
			return pq.initialEntitlement(f);
		}
		//water rights adopted from pumping quotas
		else{
			return pq.waterQuota(f,elp);
		}
	}
	
	public double aepArea(Farmer f, EvaluatedLandUsePattern elp, int year){
		
		double a;
		//before 1993 no AEP
		if(year<1993){
			a = 0.0;
		}
	
		else if(year<=1997){
		//first phase before 1997
		//vineyards not eligible for aid
		//fallow land is assumed not eligible for aid (but may be included for waterQuota)
		//area for which aid is claimed is assumed to be
		//f.sizeOfFarm - vineyards - fallow
		
			boolean noVineAndFallow[] = new boolean[Crop.NR_CROPS];
			for(int i=0; i<noVineAndFallow.length;i++){
				if((i!= Crop.VINEYARDS) && (i!=Crop.FALLOW)){
					noVineAndFallow[i] = true;
				}
			}
			a = elp.pattern.areaOfLandUses(true, noVineAndFallow)+elp.pattern.areaOfLandUses(false, noVineAndFallow);
		}
		else if(year < 2003){
			//since 1998 olives are also excluded
			boolean noVOF[] = new boolean[Crop.NR_CROPS]; //no Vine, Olives, Fallow
			for(int i=0; i<noVOF.length;i++){
				if((i!= Crop.VINEYARDS) && (i!=Crop.FALLOW) && (i!=Crop.OLIVES)){
					noVOF[i] = true;
				}
			}
			a = elp.pattern.areaOfLandUses(true, noVOF)+elp.pattern.areaOfLandUses(false, noVOF);		
		}
		else{
			//after 2003 AEP Phase 2
			//vineyards and olives are eligible again
			//AEP and Pumping Quotas get coupled: AEP aid can be claimed maximum for area for which water entitlements exist according to pumping quotas
			
			boolean noFallow[] = new boolean[Crop.NR_CROPS];
			for(int i=0; i<noFallow.length;i++){
				if(i!=Crop.FALLOW){
					noFallow[i] = true;
				}
			}
			//max AEP area is all area of farm except fallow land
			double maxAEPArea = elp.pattern.areaOfLandUses(true, noFallow)+elp.pattern.areaOfLandUses(false, noFallow);
			
			double entitledArea = pq.entitledArea(f, elp);
			
			a=  Math.min(maxAEPArea,entitledArea);
		}
		return a;
	}

}
