package model.rules;

import java.util.HashMap;

import model.*;
import model.decisionMaking.*;

public class CapCompPayments extends Rule {

	
	protected int baseYear = 1991;	//before this year, no subsidies, first entry in sub-arrays refer to 1991
	
	protected double rainfedTradCerealSubs[] = {0.0,0.0,38.0, 75.0,114.0,114.0,114.0,114.0,114.0,123,132,132};
	protected double irrigatedTradCerealSubs[] = {0.0,0.0,75.0, 151.0,228.0,228.0,228.0,228.0,228.0,247,265,265};
	protected double highYieldCerealSubs[] = {0.0, 0.0, 109.0, 219.0,331.0,331.0,331.0,331.0,331.0,358,384,384};
	protected double sunflowerSubs[] = {0.0,198.0,198.0,198.0,198.0,198.0,198.0,198.0,198.0,172,152,132};
	protected double setAsideSubs[] = {0.0,145,145,145,145,145,145,145,145,123,132,132};
	
	protected HashMap<Farmer, Double> eligibleLand;
	
	protected int rtc=0, itc=1, hc=2, sf =3, sa =4;	//indices for array
	
	protected double simplifiedSchemeThreshold = 43.8; //ha, if COP-area is bigger, simplified scheme can not be applied
	
	
	public void prepareNextStep(){	
		if(getYear() == 1992){
			yc.setXLabel("DP");
		}
	}
	
	public void evaluateLandUsePattern(EvaluatedLandUsePattern elp, Farmer f) {
		evaluateLandUsePattern(elp,f,getYear());
	}
	
	public void evaluateLandUsePattern(EvaluatedLandUsePattern elp, Farmer f, int year) {
		
		//compensatory payments not in place before 1992
		if(getYear()>=1992){	
			LandUsePattern lp = f.landUsePattern();
			double eLand=0.0;
			//set eligible land of each farmer. In year 1992, f.landUsePattern is what he cropped in Dec 1991
			if(getYear() == 1992){
				if(eligibleLand==null){
					eligibleLand = new HashMap<Farmer,Double>(GlobalVariables.nrFarmersModel());
	
				}
				LandUse l;
				for(int i=0;i<lp.length();i++){
					l = lp.get(i);
					//all land is eligible land, except tree crops
					if(Crop.cropClassFor( ((Farming)l.option).getCrop() )!=Crop.TREECROP){
						eLand+=l.size;
					}
				}
				eligibleLand.put(f, new Double(eLand));
			}
			
			////////////////////////////////
			//Calculate subisdies for potential next LandUsePattern elp.pattern
			lp= elp.pattern;
			
			//	period in [0,4], [1985-1991]->0, 1992->1, 1993->2, 1994->3, 1995->4
			int y = Math.min(sunflowerSubs.length-1,Math.max(0,year-baseYear));
			
			//collect areas of COPs
			double cropAreas[] = new double[5];
			LandUse l;
			Farming o;
			for(int i=0; i<lp.length(); i++){
				l = lp.get(i);					//error if non-farming options are included!
				o =((Farming)l.option);
				switch(o.getCrop().getType()){
					case Crop.TRADITIONALCEREALS: 
						if(o.isIrrigated()){
							cropAreas[itc]+=l.size; 
						}
						else{
							cropAreas[rtc]+=l.size;
						}
						break;
					case Crop.HIGHYIELDCEREALS: cropAreas[hc]+=l.size;break;
					case Crop.SUNFLOWER: cropAreas[sf]+=l.size;break;
					case Crop.FALLOW: cropAreas[sa] += l.size;break;
				}
			}
			double totalArea =cropAreas[rtc]+cropAreas[itc]+cropAreas[hc]+cropAreas[sf]+cropAreas[sa];
	
			//check whether more than eligible land
			eLand = eligibleLand.get(f).doubleValue();
			if(totalArea>eLand){
				totalArea = eLand;
			}
		
			//check whether simplified scheme 
			if(totalArea <= simplifiedSchemeThreshold){
				elp.capSimplifiedScheme = true;
				elp.capCompPayments = simplifiedScheme(cropAreas, y);
			}
			else{
				elp.capSimplifiedScheme = false;
				elp.capCompPayments = generalScheme(totalArea,cropAreas,y);
			}
		}
	}
	
	protected double simplifiedScheme(double[] area, int p){
		//max set-aside < COP area
		if(area[sa] > area[rtc]+area[itc]+area[hc]+area[sf]){
			return 0.0;
		}
		return (area[rtc]+area[itc]+area[hc]+area[sf]+area[sa])*rainfedTradCerealSubs[p];
		
	}
	protected double generalScheme(double sumArea, double[] cropAreas, int p){
		
		double setAsideObligation = 0.1;
		
		//criteria missed -> no subsidies (either less than setAsideObligation or less than half of eligible land is actually used)
		if((cropAreas[sa] < setAsideObligation*sumArea)|| (cropAreas[rtc]+cropAreas[itc]+cropAreas[hc]+cropAreas[sf]<0.5*sumArea)){
			return 0.0;
		}
		//Sum up areas, starting with most profitable ones, respect set-aside obligation
		double a = sumArea*0.9; //land to be allocated most profitably
		//hy cerals
		cropAreas[hc] = Math.min(cropAreas[hc],a);
		a -= cropAreas[hc];
		//irrigated trad cereals
		cropAreas[itc] =Math.min(cropAreas[itc],a);
		a -= cropAreas[itc];
		//sunflower
		cropAreas[sf] =Math.min(cropAreas[sf],a);
		a -= cropAreas[sf];
		//set-aside (include here obligatory set-aside to max possible area)
		a+=+sumArea*0.1;
		cropAreas[sa] =Math.min(cropAreas[sa],a);
		a -= cropAreas[sa];
		//trad rainfed cereals
		cropAreas[rtc] =Math.min(cropAreas[rtc],a);
		a -= cropAreas[rtc]; //should be zeor by now....
		
		//start with hy-cereals
		double subs = cropAreas[rtc]*rainfedTradCerealSubs[p]+cropAreas[itc]*irrigatedTradCerealSubs[p]+cropAreas[hc]*highYieldCerealSubs[p]+
			cropAreas[sf]*sunflowerSubs[p]+cropAreas[sa]*setAsideSubs[p];
		return subs;
	}
}
