package model.rules;

import java.util.HashMap;

import model.*;
import model.decisionMaking.EvaluatedLandUsePattern;

public class PumpingQuotas extends Rule {



	public double efficiencyOfControl = 0.01; 	//0.0-> no control, 1.0->perfect control 
	
	protected HashMap<Farmer,LandUsePattern> referenceState;
	
	public void prepareNextStep(){	
		if(getYear() == 1986){ //water law active 	
			yc.setXLabel("WL");
		}
		if(getYear() == 1991){	
			yc.setXLabel("PQ");
		}
	}
	
	
	public void evaluateLandUsePattern(EvaluatedLandUsePattern elp, Farmer f) {
		
//		Water law not yet introduced
		if(getYear() < 1985){
			elp.complyWithWaterLaw = true;
		}
		
		//collect water rights for farmers and memorize in hashmap 
		else if(getYear() == 1985){
			
			if(referenceState == null){ //no information for any farmer stored yet
				referenceState = new HashMap<Farmer,LandUsePattern>(1000);
			}
			referenceState.put(f, LandUsePattern.clone(f.landUsePattern()));
			//water law still not active
			elp.complyWithWaterLaw = true;
		}
		else{
			//1985 onwards
			//check if farmer subscribes to AEP which overwrites pumping quotas (?)
			if(elp.aepOption != AEP.NO_AEP){
				elp.complyWithWaterLaw = true;
			}
			else{
				//No AEP -> check if water quotas are respected
				double legalWater = waterQuota(f,elp);
				
				//violation of law
				if(elp.pattern.waterUsed()>legalWater){
					elp.complyWithWaterLaw = false;
					//CHANGE: set the double WaterOverUse
					elp.waterOverUse = 1-(legalWater / elp.pattern.waterUsed()); 
					//water law imposes a fine of 7cent/m^3 abstracted illegally.
					//Assume rational calculation of risk (using efficiencyOfControl
					elp.waterLawPenalties = (elp.pattern.waterUsed() -legalWater)*0.07*efficiencyOfControl;
				}
				//compliance with law
				else{
					elp.complyWithWaterLaw = true;
				}	
			}
		}
	}
	
	public double waterQuota(Farmer f, EvaluatedLandUsePattern elp){
		

		double vineQuota = 2000;
		double quota1985 = 4278; //m^3/ha
		
//		1986-1990
		if(getYear()<1991){
			//calculate water rights: 
			//max claim: initial entitlement: irrigated area in 1985 excluding vineyards and Fallow
			//max claim: current area without vineyards (for which no water may be claimed)
			//quota per ha is constant
						
			return entitledArea(f,elp)*quota1985;
		}
		//1991 onwards
		else{
			
			double a = entitledArea(f,elp);	//area for which irrigation rights can be claimed
			double quota;
			if(getYear()<1997){ //vineyards are not included
				quota= quota1985;
			}
			else{ //since 1997, vineyards may be irrigated and are thus
				//included in calculation, however they have a lower quota,
				//which is 2000m^3/ha (compared to 4278 m^3/ha)
				//therefore include first of all non-vine area
				
				//get vine area of this farmer
				boolean vine[] = new boolean[Crop.NR_CROPS];
				vine[Crop.VINEYARDS] = true;
				double vineArea = elp.pattern.areaOfLandUses(true, vine) + elp.pattern.areaOfLandUses(false, vine);

				//get regular quota for as big area as possible
				//calculate non-vine area that can be claimed, complement by vine-area
				//Residual quota calculated from relation non-vine/vine area
				double claimedNoVineArea = Math.min(a,elp.pattern.size()-vineArea);			
				quota = (claimedNoVineArea*quota1985+vineQuota*(a-claimedNoVineArea))/a;		
			}
			
			//<5 ha: 100%=4278m^3/ha (for normal quota)
			double waterGranted = Math.min(a, 5)*quota;
			a -=5;
			//5-10ha: 50%=2140m^3/ha
			if(a>0.0){ 
				waterGranted+=Math.min(a,5)*quota*0.5;
				a-=5;
			}
			//10-30ha: 35%=1497m^3/ha
			if(a>0.0){
				waterGranted+=Math.min(a,20)*quota*0.35;
				a-=20;
			}
			//>30ha: 25%=1069m^3/ha
			if(a>0.0){
				waterGranted+=a*quota*0.25;
			}
			return waterGranted;
			
		}
	}
	
	
	//the area for which a farmer may claim water (maximum is 'reference area' but may be
	//less, e.g. if later on vineyards have been extended, for which <1997 no water may be 
	//claimed
	public double entitledArea(Farmer f, EvaluatedLandUsePattern elp){
		
		double a;
		
		double refArea = getReferenceArea(f);
		
//		1986-1996
		if(getYear()<1997){
			//calculate water rights: 
			//max claim: initial entitlement: irrigated area in 1985 excluding vineyards
			//max claim: current area without vineyards (for which no water may be claimed)
			//quota per ha is constant
			
			boolean noVine[] = new boolean[Crop.NR_CROPS];
			for(int i=0; i<noVine.length;i++){
				if(i!= Crop.VINEYARDS){
					noVine[i] = true;
				}
			}
			double currentArea = elp.pattern.areaOfLandUses(true, noVine) + elp.pattern.areaOfLandUses(false, noVine);
			
			a = Math.min(refArea,currentArea);
		}
		//1997 onwards: vineyards are included
		else{

				a = refArea;		
		}
		return a;
	}
	
	public double initialEntitlement(Farmer f){
		double quota1985 = 4278; //m^3/ha
		return getReferenceArea(f)*quota1985;
	}
	//area which was irrigated by this farmer as memorized in the referenceState (excluding vine and fallow)
	public double getReferenceArea(Farmer f){

		double refArea = 0.0;
		
		if(referenceState != null){
			
			boolean noVineAndFallow[] = new boolean[Crop.NR_CROPS];
			for(int i=0; i<noVineAndFallow.length;i++){
				if((i!= Crop.VINEYARDS) && (i!=Crop.FALLOW)){
					noVineAndFallow[i] = true;
				}
			}
			refArea = referenceState.get(f).areaOfLandUses(true, noVineAndFallow);

		}
		return refArea;
	}
}
